/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
/****************************************************************************
 * This file contains basic functions 
 * - to allocate storage
 * - to set values
 * - conversion functions
 * - to manipulate strings
 * - UNIX specific utilities
 * - to report errors
 * used in the ERGO-shell
 ****************************************************************************/

#include <stdio.h>

#ifdef __GNUC__
#ifdef __sparc__
#include <gcc-include/varargs.h>
#else
#include <varargs.h>
#endif __sparc__
#else
#include <varargs.h>
#endif __GNUC__

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/dir.h>
#include <sys/file.h>
#include <sys/time.h>
#include <strings.h>
#include <ctype.h>
#include <errno.h>
#include <exBasics.h>
#include <exErrors.h>

extern char *asctime();
extern int errno;

void fatal();

/****************************************************************************
 * functions to allocate memory
 ***************************************************************************/

/* Malloc ********************************************************************
 * allocates 'size' bytes of memory initialized with zeros to returned pointer
 *****************************************************************************/

char *Malloc(size)
unsigned size;
{ 
  char *ptr;
  
  if (ptr = (char *) malloc(size))
    bzero(ptr, size);
  else
    fatal("cannot allocate storage","Malloc");

  return(ptr);
}

/* Calloc *********************************************************************
 * allocates 'size' bytes of initialized storage to each of 'elems' elements
 *****************************************************************************/

char **Calloc(elems,size)
unsigned elems, size;
{
  char **ptr;
  
  if ((ptr = (char **) calloc(elems,size)) == (char **) 0)
      fatal("cannot allocate storage","Calloc");

  return(ptr);
}

/****************************************************************************
 * functions to set values
 ***************************************************************************/

/* setMask: ******************************************************************
 * fill an integer array with a given integer
 ****************************************************************************/

void setMask(m, cnt, setreset)
     int m[];
     int cnt;
     int setreset;
{
  register i;

  for (i = 0; i < cnt; i++)
    m[i] = setreset;
}

/* setBit: *******************************************************************
 * sets the required bit in the integer parameter 'mask' to 1 or 0
 ****************************************************************************/

void setBit(mask, bit, setreset)
     int *mask, bit, setreset;
{
  switch(setreset) {
  case 0:
    *mask &= ~(1 << bit);
    break;
  case 1:
    *mask |= (1 << bit);
    break;
  }
}

/*****************************************************************************
 * conversion functions
 ****************************************************************************/

/* convTime: *****************************************************************
 * converts a time given as long integer (seconds since 1.1.1970) to an 
 * ascii-string; the time is corrected for day-light savings and local time 
 * zone
 ****************************************************************************/

char *convTime(clock)
     long *clock;
{
  return(asctime(localtime(clock)));
}

/****************************************************************************
 * functions to manipulate strings
 ***************************************************************************/

/* strCopy *******************************************************************
 * copy a string by allocating space and then copying 
 *****************************************************************************/

char *strCopy(str)
     char *str;
{
  char *s;

  if (! str)
    return((char *) 0);

  s = (char *) malloc(strlen(str) + 1);
  if (strlen(str) <= 0)
    *s = '\0';
  else
    strcpy(s,str);
  return(s);
}

/* strCmp ********************************************************************
 * compare strings; return TRUE if differences occur or one is a nullstring
 ****************************************************************************/

Boolean strCmp(str1, str2)
     char *str1, *str2;
{
  if (str1 != (char *) 0) {
    if (str2 == (char *) 0)
      return(True);
    else if (strcmp(str1, str2) != 0)
      return(True);
  }
  else if (str2 != (char *) 0)
    return(True);

  return(False);
}

/* emptyStr ****************************************************************
 * Checks if a string only consists of space, tab, carriage return, newline,
 * vertical tab or formfeed characters; returns TRUE in this case
 ****************************************************************************/

Boolean emptyStr(string)
     char *string;
{
  if (!string)
    return(True);

  while (*string) {
    if (isspace(*string) == 0)
      return(False);
    string++;
  }

  return(True);
}

/* charInSet *****************************************************************
 * return TRUE if character c is in string str
 *****************************************************************************/

Boolean charInSet(c, str)
     char c;
     char *str;
{
  while (*str) {
    if (*str == c)
      return(True);
    str++;
  }
  return(False);
}

/* cntCharInStr *************************************************************
 * counts occurence of character in string
 ****************************************************************************/

cntCharInStr(chr, str)
     char chr, *str;
{  
  int count = 0;
  char *index();

  if (strlen(str) == 0)
    return(count);

  if (*str == chr)    
    count++;

  while ((str = index(str + 1, chr)) != NULL)
    count++;

  return(count);
} 

/* makeStr *******************************************************************
 * Make a null-terminated string of the chars between two pointers pointing to
 * different positions in a string (last character is the character prior to 
 * second pointer)
 *****************************************************************************/

char *makeStr(start, end, size)
     char *start;             /* pointer to first char */
     char *end;               /* pointer to first position behind last char */
     int size;                /* size of buffer containing result string */
{
  char *buf = Malloc(size);
  char *p = buf;
 
  while (start < end && strlen(buf) < size)
    *p++ = *start++;

  if (start != end) {
    setError(ERR_MAX);
    return((char *) NULL);
  }

  *p = EOS;
  return (buf);
}

/* parseStr *****************************************************************
 * Make a null-terminated string of the characters in between two specified 
 * characters; dates 'pat' up 
 *****************************************************************************/

char *parseStr(pat, char1, char2, size)
     char **pat;
     char char1, char2;
     int size;                    /* size of resulting string */
{
  char *n, *m;

  if (!(n = index(*pat, char1)))
    return ((char *) 0);

  if (!(m = index(n, char2)))
    return ((char *) 0);

  *pat = m;

  return(makeStr(n + 1, m, size));
}

/* makeStringList *************************************************************
 * generate a list of tokens from a string of tokens separated by del; the
 * number of tokens is returned;
 *****************************************************************************/

int makeStringList(buffer, list, cnt, del)
     char *buffer, **list;
     int cnt;
     char del;
{
  register i;
  register char *ptr1, *ptr2;

  ptr1 = buffer;
  ptr2 = ptr1;

  while (*ptr1 == del)
    ptr1++;

  for (i = 0; i < cnt && ptr2 && *ptr2; ) {
    if (ptr2 = index(ptr1, del)) {
      *ptr2++ = EOS;
      while (*ptr2 == del)		/* skip multiple occurrences of del */
	ptr2++;
      if (*ptr1) 
	list[i++] = strCopy(ptr1);
      ptr1 = ptr2;
    } else {
      list[i++] = strCopy(ptr1);
    }
  }
  return(i);
}

/* sortStringList: ***********************************************************
 * sort a list consisting of string using the qsort-routine
 ****************************************************************************/

sortStringList(list, cnt)
     char **list;
     int cnt;
{
  static int qcompare();

  qsort(list, cnt, sizeof(char *), qcompare);
}

/* qcompare: ****************************************************************
 * function used for comparison of strings to be used with function qsort
 * within function sortStringList.
 ****************************************************************************/

static int qcompare(s1, s2)
     char **s1;
     char **s2;
{
  return(strcmp(*s1, *s2));
}

/* writeString: *************************************************************
 * write a string as a tuple consisting of a int representing the length
 * followed by the actual characters; the string can be reread by 'readString'
 ****************************************************************************/

writeString(fd,string)
int fd;
char *string;
{
  int len;
  if (! string) {
    len = 0;
    Write(fd,&len,sizeof(int));
  } else {
    len = strlen(string);
    Write(fd,&len,sizeof(int));
    Write(fd,string,len);
  }
  return(OK);
}

/* readString: ***************************************************************
 * read a string that consists of a tuple of an int representing the length
 * followed by the actual characters; used to read strings written by 
 * 'writeString'
 ****************************************************************************/

readString(fd, string)
     int fd;
     char **string;
{
  int len;

  read(fd, &len, sizeof(int));
  if (len < 0 || len > EX_MAXSTR) {
    return(setError(ERR_MAX));
  }
  if (len == 0) {
    *string = (char *) 0;
    return(OK);
  }
  *string = (char *) malloc(len + 1);
  Read(fd,*string,len);
  *(*string + len) = '\0';
  return(OK);
}

/* fformatText: ***************************************************************
 * formats a text that it fits within a given length; it returns the number of
 * lines that are separated by cr`s. Word separators are ",.?;!" and blank.
 ****************************************************************************/

fformatText(text, length, ftext)
     char *text;			/* input string */
     int length;			/* length required */
     char **ftext;			/* formated output string */
{
  int newlen = strlen(text) + 2 * (strlen(text) / length);
  register char *p0, *p1, *ft;
  char *ftp;
  char *sep = ",.;?! ";			/* the string of separators */
  int pos, lastPos;
  int i, lines = 1;

  if (strlen(text) < length) {		/* fits into one line */
    *ftext = Malloc(strlen(text) + 2);
    sprintf(*ftext, "%s\n", text);
    return(1);
  }

  ft = ftp = Malloc(newlen);

  p0 = text;
  p1 = text;
  pos = 0;

  while (*p1) {
    if (charInSet(*p1, sep))
      lastPos = pos;
    if (pos >= length) {		/* line break */
      for (i = 0; i <= lastPos; i++)
	*ft++ = *p0++;
      *ft++ = '\n';
      pos -= lastPos;
      lastPos = pos;
      while (*p1 && *p1 == ' ') 	/* remove duplicate blanks */
	p1++;
      lines++;
    } else 
	pos++;

    p1++;
  }

  while (*p0)
    *ft++ = *p0++;

  if (*(ft - 1) != '\n') {
    *ft++ = '\n';
    *ft = EOS;
  }

  *ftext = ftp;
  return(lines);			/* return the number of lines */
}

/****************************************************************************
 * UNIX-specific utilities
 ***************************************************************************/

/* expand_name ****************************************************************
 * expand ~ in file name (HOME directory)
 *****************************************************************************/

expand_name(n, len)
     char *n;
     int len;
{
  char *tmp = Malloc(len + 1);
  
  if (! ((*n == '~') && (n[1] == '/'))) {
    setError(ERR_NOEXIST);
    return(ERROR);
  }

  strncpy(tmp, getenv("HOME"), len);
  strncat(tmp, n + 1, len - strlen(tmp));
  strncpy(n, tmp, len);
  return(OK);
}

/* makeFullName: *************************************************************
 * generates a path name from the two components 'dir' and 'file' and copies 
 * the result into the user supplied buffer 'fullName' that is expected to 
 * have the size given in 'len'
 *****************************************************************************/

makeFullName(dir, file, fullName, len)
     char *dir, *file, fullName[];
     int len;
{
  if (! dir ) {
    if (strlen(file) < len) {
      strcpy(fullName, file);
      return(setError(ERR_NOERR));
    } else     
      return(setError(ERR_MAX));
  }
      
  if ((strlen(dir) + strlen(file)) >= len)
    return(setError(ERR_MAX));

  strcpy(fullName, dir);
  if ((fullName[strlen(fullName) - 1] != '/') && (file[0] != '/'))
    strcat(fullName, "/");

  strcat(fullName, file);
  return(setError(ERR_NOERR));
}

/* fileExists: ***************************************************************
 * check whether the file specified by the given name exists in the current
 * directory.
 ****************************************************************************/

Boolean fileExists(file)
     char *file;		/* name of the file to be checked */
{
  int fd;

  if ((fd = open(file, O_CREAT | O_EXCL)) == -1) {
    if (errno == EEXIST)
      return(True);
  } else {
    close(fd);
    unlink(file);
  }

  return(False);
}

/* getFileSize: **************************************************************
 * get the size of a file via the stat system call.
 ****************************************************************************/

int getFileSize(file)
     char *file;
{
  struct stat sbuf;

  if (stat(file, &sbuf) < 0) {  		/* get file size */
    setError(ERR_READ);
    return(-1);
  }
  
  return(sbuf.st_size);
}
  
/* touchFile: *****************************************************************
 * touch the specified file, i.e. set its accessed and updated times 
 * (see utimes(2)).
 *****************************************************************************/

int touchFile(file, a_time, u_time)
     char *file;
     long a_time, u_time;
{
  struct timeval tvp[2];

  tvp[0].tv_sec = a_time;
  tvp[0].tv_usec = 0L;

  tvp[1].tv_sec = u_time;
  tvp[1].tv_usec = 0L;

  return(utimes(file, tvp));
}

/* loadDirectory: ************************************************************
 * opens a directory and stores all file-names, except "." and ".." into the 
 * provided array.
 ****************************************************************************/

int loadDirectory(directory, table, size)
     char *directory;
     char *table[];
     int size;
{
  DIR *dirp;
  register struct direct *entry;
  register cnt = 0;

  if ((dirp = opendir(directory)) == (DIR *) NULL) {
    setError(ERR_OPEN);
    return(-1);
  }

  for (entry = readdir(dirp); entry && cnt < size; entry = readdir(dirp)) {
    if (strcmp(entry->d_name, ".") == 0 || strcmp(entry->d_name, "..") == 0)
      continue;

    table[cnt++] = strCopy(entry->d_name);
  }

  closedir(dirp);
  return(cnt);
}

/* cleanPath: ****************************************************************
 * 'cleans' a string that denotes a UNIX file-path; eliminates and handles 
 * '.' and '..' components; the cleaned string is copied to the buffer 'new';
 * size gives the size of buffer 'new'
 ****************************************************************************/

cleanPath(new, old, size)
     char *new, *old;
     int size;
{
  int len, cnt;
  char *list[25], *nlist[25];
  int nl, i;
  
  if (strcmp(old, "/") == 0) {
    *new= '/';
    return(OK);
  }

  cnt = makeStringList(old, list, 25, '/');	/* break into components */
  new[0] = EOS;

  for (i = 0, nl = 0; i < cnt; i++) {
    if (strcmp(list[i], ".") == 0)		/* skip "." */
      continue;
    if (strcmp(list[i], "..") == 0) {		/* recognized ".." */
      if (nl <= 0)
	return(-1);
      else
	nl--;					/* go back one level */
    } else					/* store current level */
      nlist[nl++] = list[i];
  }
  
  
  if (*old == '/')
    strcpy(new, "/");
  else 
    *new = EOS;

  for (i = 0, len = 0; i < nl; i++) {
    if (! nlist[i][0])
      continue;
    len = len + strlen(nlist[i]) + 1;
    if (len >= size) {
      setError(ERR_MAX);
      return(ERROR);
    }

    strcat(new, nlist[i]);
    strcat(new, "/");
  }

  return(OK);
}

/* isDirFile: ***************************************************************
 * returns TRUE if the file is a directory
 ****************************************************************************/

Boolean isDirFile(file)
     char *file;
{
  struct stat sbp;
  char msg[128];

  if ((stat(file, &sbp)) == -1) {
    sprintf(msg, "Internal error: cannot fstat file -- %s --\n", file);
    exError("isDirFile", msg, NULL);
    setError(ERR_READ);
    return(False);
  }
  return(((sbp.st_mode & S_IFMT) == S_IFDIR) ? True : False);
}

/* exWait: *******************************************************************
 * wait for a certain time: ticks gives the time in steps of 50 ms
 ****************************************************************************/

exWait(ticks)
     int ticks;
{
    static struct timeval flashPeriod = {0, 50000}; /* 50 ms */
    extern int select();
    register i;
    
    for (i = 0; i < ticks; i++)
      select(0, (fd_set *) 0, (fd_set *) 0, (fd_set *) 0, &flashPeriod);
}

/* fatal: ********************************************************************
 * print out error list and exit program
 ****************************************************************************/
/*VARARGS*/
void fatal(va_alist)
     va_dcl
{
  char *s1, *s2;
  va_list ap;

  va_start(ap);

  s1 = va_arg(ap, char *);
  s2 = va_arg(ap, char *);
  fprintf((FILE *)stderr, "%s (--> proc <%s>)\n", s1, s2);

  while (s1 = va_arg(ap, char *))
    fprintf(stderr, "%s\n", s1);

  va_end(ap);
  exit(1);
}

/* exError: *****************************************************************
 * print out error list
 ****************************************************************************/
/*VARARGS*/
void exError(va_alist)
     va_dcl
{
  char *s1;
  char *s2;
  va_list ap;

  va_start(ap);

  s1 = va_arg(ap, char *);
  s2 = va_arg(ap, char *);

  fprintf((FILE *)stderr,"%s: %s", s1, s2);

  while (s1 = va_arg(ap, char *))
    fprintf(stderr, "%s", s1);

  fprintf(stderr, "\n");

  va_end(ap);
}

/* power: ********************************************************************
 * calculate the power of x to y
 ****************************************************************************/

int power(x, y)
     register x, y;
{
  register res = x;

  if (y == 0)
    return(1);

  while (--y > 0)
    res = res * x;

  return(res);
}






